@@ -7,9 +7,9 @@
 //! subgroup of order `2^n` for some `n`.
 
 use crate::{
-    fp::{FP126, FP32, FP64, FP96},
+    codec::{CodecError, Decode, Encode},
+    fp::{FP128, FP32, FP64, FP96},
     prng::{Prng, PrngError},
-    vdaf::suite::Suite,
 };
 use serde::{
     de::{DeserializeOwned, Visitor},
@@ -41,6 +41,9 @@ pub enum FieldError {
     /// Error while performing I/O.
     #[error("I/O error")]
     Io(#[from] std::io::Error),
+    /// Error encoding or decoding a field.
+    #[error("Codec error")]
+    Codec(#[from] CodecError),
 }
 
 /// Byte order for encoding FieldElement values into byte sequences.
@@ -78,6 +81,8 @@ pub trait FieldElement:
     + Into<Vec<u8>>
     + Serialize
     + DeserializeOwned
+    + Encode
+    + Decode
     + 'static // NOTE This bound is needed for downcasting a `dyn Gadget<F>>` to a concrete type.
 {
     /// Size in bytes of the encoding of a value.
@@ -119,30 +124,16 @@ pub trait FieldElement:
     /// # Errors
     ///
     /// An error is returned if the provided slice is too small to encode a field element or if the
-    /// result encodes an integer larger than the field modulus.
+    /// result encodes an integer larger than or equal to the field modulus.
     ///
     /// # Warnings
     ///
     /// This function should only be used within [`prng::Prng`] to convert a random byte string into
-    /// a field element. Use [`Self::try_from_reader`] to deserialize field elements. Use
+    /// a field element. Use [`Self::decode`] to deserialize field elements. Use
     /// [`field::rand`] or [`prng::Prng`] to randomly generate field elements.
     #[doc(hidden)]
     fn try_from_random(bytes: &[u8]) -> Result<Self, FieldError>;
 
-    /// Interprets the next [`Self::ENCODED_SIZE`] bytes from the input slice as an element of the
-    /// field.
-    ///
-    /// # Errors
-    ///
-    /// An error is returned if the provided slice is too small to encode a field element or if the
-    /// result encodes an integer larger than the field modulus.
-    ///
-    /// # Notes
-    ///
-    /// Ideally we would implement `TryFrom<R: Read> for FieldElement` but the stdlib's blanket
-    /// implementation of `TryFrom` forbids this: <https://github.com/rust-lang/rust/issues/50133>
-    fn try_from_reader<R: Read>(reader: &mut R) -> Result<Self, FieldError>;
-
     /// Returns the size of the multiplicative subgroup generated by `generator()`.
     fn generator_order() -> Self::Integer;
 
@@ -181,7 +172,7 @@ pub trait FieldElement:
     ///
     /// Returns an error if the length of the provided byte slice is not a multiple of the size of a
     /// field element, or if any of the values in the byte slice are invalid encodings of a field
-    /// element as documented in [`Self::try_from_reader`].
+    /// element, because the encoded integer is larger than or equal to the field modulus.
     ///
     /// # Notes
     ///
@@ -194,7 +185,7 @@ pub trait FieldElement:
         }
         let mut vec = Vec::with_capacity(bytes.len() / Self::ENCODED_SIZE);
         for chunk in bytes.chunks_exact(Self::ENCODED_SIZE) {
-            vec.push(Self::try_from_reader(&mut Cursor::new(chunk))?);
+            vec.push(Self::get_decoded(chunk)?);
         }
         Ok(vec)
     }
@@ -475,6 +466,23 @@ macro_rules! make_field {
             }
         }
 
+        impl Encode for $elem {
+            fn encode(&self, bytes: &mut Vec<u8>) {
+                let slice = <[u8; $elem::ENCODED_SIZE]>::from(*self);
+                bytes.extend_from_slice(&slice);
+            }
+        }
+
+        impl Decode for $elem {
+            fn decode(bytes: &mut Cursor<&[u8]>) -> Result<Self, CodecError> {
+                let mut value = [0u8; $elem::ENCODED_SIZE];
+                bytes.read_exact(&mut value)?;
+                $elem::try_from_bytes(&value, u128::MAX).map_err(|e| {
+                    CodecError::Other(Box::new(e) as Box<dyn std::error::Error + 'static + Send + Sync>)
+                })
+            }
+        }
+
         impl FieldElement for $elem {
             const ENCODED_SIZE: usize = $encoding_size;
             type Integer = $int;
@@ -497,12 +505,6 @@ macro_rules! make_field {
                 $elem::try_from_bytes(bytes, $fp.bit_mask)
             }
 
-            fn try_from_reader<R: Read>(reader: &mut R) -> Result<Self, FieldError> {
-                let mut bytes = [0u8; $elem::ENCODED_SIZE];
-                reader.read_exact(&mut bytes)?;
-                $elem::try_from_bytes(&bytes, u128::MAX)
-            }
-
             fn generator() -> Self {
                 Self($fp.g)
             }
@@ -567,10 +569,10 @@ make_field!(
 );
 
 make_field!(
-    /// `GF(85070591730234613043491808580380655617)`, a 126-bit field.
-    Field126,
+    /// `GF(340282366920938462946865773367900766209)`, a 128-bit field.
+    Field128,
     u128,
-    FP126,
+    FP128,
     16,
     ByteOrder::BigEndian,
 );
@@ -619,9 +621,7 @@ pub(crate) fn split_vector<F: FieldElement>(
 
 /// Generate a vector of uniform random field elements.
 pub fn random_vector<F: FieldElement>(len: usize) -> Result<Vec<F>, PrngError> {
-    Ok(Prng::generate(Suite::Aes128CtrHmacSha256)?
-        .take(len)
-        .collect())
+    Ok(Prng::new()?.take(len).collect())
 }
 
 #[cfg(test)]
@@ -630,7 +630,6 @@ mod tests {
     use crate::fp::MAX_ROOTS;
     use crate::prng::Prng;
     use assert_matches::assert_matches;
-    use std::io::{Cursor, Write};
 
     #[test]
     fn test_endianness() {
@@ -665,7 +664,7 @@ mod tests {
     // experimental" https://github.com/rust-lang/rust/issues/15701
     #[allow(clippy::eq_op)]
     fn field_element_test<F: FieldElement>() {
-        let mut prng: Prng<F> = Prng::generate(Suite::Aes128CtrHmacSha256).unwrap();
+        let mut prng: Prng<F, _> = Prng::new().unwrap();
         let int_modulus = F::modulus();
         let int_one = F::Integer::try_from(1).unwrap();
         let zero = F::zero();
@@ -738,15 +737,13 @@ mod tests {
         // serialization
         let test_inputs = vec![zero, one, prng.get(), F::from(int_modulus - int_one)];
         for want in test_inputs.iter() {
-            println!("check {:?}", want);
-            let mut bytes: Vec<u8> = vec![];
-            bytes.write_all(&(*want).into()).unwrap();
+            let mut bytes = vec![];
+            want.encode(&mut bytes);
 
             assert_eq!(bytes.len(), F::ENCODED_SIZE);
 
-            let got = F::try_from_reader(&mut Cursor::new(&bytes)).unwrap();
+            let got = F::get_decoded(&bytes).unwrap();
             assert_eq!(got, *want);
-            assert_eq!(bytes.len(), F::ENCODED_SIZE);
         }
 
         let serialized_vec = F::slice_into_byte_vec(&test_inputs);
@@ -775,7 +772,7 @@ mod tests {
     }
 
     #[test]
-    fn test_field126() {
-        field_element_test::<Field126>();
+    fn test_field128() {
+        field_element_test::<Field128>();
     }
 }
