@@ -2,22 +2,21 @@
 // SPDX-License-Identifier: MPL-2.0
 
 #![warn(missing_docs)]
+#![cfg_attr(docsrs, feature(doc_cfg))]
 
 //! Libprio-rs
 //!
 //! Implementation of the [Prio](https://crypto.stanford.edu/prio/) private data aggregation
 //! protocol. For now we only support 0 / 1 vectors.
 
-#[macro_use]
-extern crate static_assertions;
-
 pub mod benchmarked;
 pub mod client;
+pub mod codec;
 pub mod encrypt;
 mod fft;
 pub mod field;
+pub mod flp;
 mod fp;
-pub mod pcp;
 mod polynomial;
 mod prng;
 pub mod server;
